/**
  *  \file client/dialogs/prefdlg.cc
  *  \brief Preferences dialog
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include <cstring>
#include "client/dialogs/prefdlg.h"
#include "client/chart.h"
#include "client/dialogs/chartconfig.h"
#include "client/markerconfig.h"
#include "client/widgets/configedit.h"
#include "client/widgets/configstorage.h"
#include "client/widgets/configview.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/lock.h"
#include "game/pref.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/cardgroup.h"
#include "ui/widgets/cardtab.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/radiobutton.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/tree.h"
#include "util/nls.h"
#include "client/dialogs/labeldlg.h"
#include "game/global.h"
#include "game/team.h"
#include "cpluslib/misc.h"
#include "ui/util.h"
#include "ui/richparser.h"
#include "ui/msgbox.h"
#include "game/host.h"
#include "game/config.h"
#include "client/dialogs/passwordchange.h"
#include "client/dialogs/racename.h"
#include "ui/widgets/quit.h"

namespace {
    class WConfigBackupEditor : public WConfigEditor::Editor {
     public:
        WConfigBackupEditor(Config& cfg);
        string_t toString(string_t arg);
        void edit(string_t arg);

     private:
        Config& cfg;
    };

    class WConfigBackupEditorWindow : public UIWindow {
     public:
        WConfigBackupEditorWindow(ConfigStringOption& opt, string_t def);
     private:
        ConfigStringOption& opt;
        string_t def;
        ObjectHolder h;
        Observable<int> selection;
        UIInputLine* input;

        void init();
        void onMove();
        void onOK();
    };

    class WConfigStarchartEditor : public WConfigEditor::Editor {
     public:
        string_t toString(string_t arg);
        void edit(string_t arg);
    };

    class WConfigBooleanEditor : public WConfigEditor::Editor {
     public:
        WConfigBooleanEditor(Config& cfg);
        string_t toString(string_t arg);
        void edit(string_t arg);

     private:
        Config& cfg;
    };

    class WConfigLabelEditor : public WConfigEditor::Editor {
     public:
        string_t toString(string_t arg);
        void edit(string_t arg);
    };

    enum EditorType {
        NoEditor,
        DefaultEditor,
        BackupEditor,
        StarchartEditor,
        BoolEditor,
        LabelEditor
    };

    static const char*const game_option_names[] = {
        N_("Set result file password..."),
        N_("Change race name..."),
        N_("Set remote control defaults..."),
        N_("Set message detail..."),
        N_("Set message language..."),
        N_("Configuration (pconfig.src)"),
        N_("Friendly code list (xtrafcode.txt)"),
        N_("Race names (race.nm)"),
    };

    enum GameOption {
        PasswordOption,
        RacenameOption,
        RemoteOption,
        FilterOption,
        LanguageOption,
        GetConfigOption,
        GetFCodeOption,
        GetRaceNameOption
    };

    class WConfigGameEditor : public WConfigEditor::Editor {
     public:
        WConfigGameEditor(GameOption option);
        string_t toString(string_t arg);
        void edit(string_t arg);

     private:
        void editFilter(GCommandContainer& cmds);
        void editRemote(GCommandContainer& cmds);
        void editLanguage(GCommandContainer& cmds);
        void toggleCommand(GCommandContainer& cmds, GCommand::Type type, int id);
        GameOption option;
    };

    /** Preferences dialog. */
    class WPreferencesDialog : public UIWindow {
     public:
        WPreferencesDialog(Config& cfg);

     private:
        void init();
        void onMove();

        Config& cfg;

        ObjectHolder h;
        UITreeWidget* settings;
        UITreeWidget* game_settings;
        WConfigView* details;
        WConfigEditor* editor;
        WConfigEditor* game_editor;
        WConfigStorageControl* storage;

        ptr_vector<WConfigEditor::Editor> editors;
        std::auto_ptr<WConfigEditor::Editor> game_editor_obj;
    };

    struct ConfigItem {
        const char* name;
        int         level;
        const char* info;
        EditorType  editor;
        const char* arg;
    };
}

static const ConfigItem config_tree[] = {
    { N_("Maketurn"),                            0, "", NoEditor,         "" },
    {   N_("Backup..."),                         1, "", BackupEditor,     "Backup.Turn,%d/backups/turn%p.%t" },
    {   N_("Upload target..."),                  1, "", DefaultEditor,    "Maketurn.Target" },
    //   TRN file format
    //   Use all message files
    //   Non-Player messages
    //   PHost identification
    //   Wait after each race
    { N_("Unpack"),                              0, "", NoEditor,         "" },
    {   N_("Data file format"),                  1, "", BoolEditor,       "Unpack.Format" },
    //   Race name changes
    //   Convert text files in UTILx.DAT
    {   N_("Create TARGETx.EXT files"),          1, "", BoolEditor,       "Unpack.TargetExt" },
    //   Source directory...
    //   Backup directory...
    {   N_("Backup..."),                         1, "", BackupEditor,     "Backup.Result,%d/backups/result%p.%t" },
    {   N_("Error correction"),                  1, "", BoolEditor,       "Unpack.FixErrors" },
    //   Wait after each race
    { N_("Program Interoperation"),              0, "", NoEditor,         "" },
    //   Winplan directory...
    //   Decompile PHost commands
    //   Save PHost commands in messages
    //   Update HCONFIG.HST
    {   N_("Rewrap incoming messages"),          1, "", BoolEditor,       "Messages.RewrapInbox" },
    //   Host 3.22.031 Torp Safety
    { N_("Preferences"),                         0, "", NoEditor,         "" },
    {   N_("User Interface"),                    1, "", NoEditor,         "" },
    //     Confirm exit from race screen
    //     Expensive graphical effects
    //     Display options...
    //     Tip of the day
    {     N_("Thousands-separator"),             2, "", BoolEditor,       "Display.ThousandsSep" },
    {     N_("Use clans everywhere"),            2, "", BoolEditor,       "Display.Clans" },
    //     Cache ship lists
    {     N_("Waypoint tied to scanner"),        2, "", BoolEditor,       "Scanner.TieWaypoint" },
    //     Scanner/Waypoint independant
    //     Display Auto Waypoints in map
    //   VCR...
    {   N_("Starchart"),                         1, "", NoEditor,         "" },
    {     N_("General Settings..."),             2, "", StarchartEditor,  "" },
    {     N_("Object Labels..."),                2, "", LabelEditor,      "" },
    {   N_("Collapse old messages"),             1, "", BoolEditor,       "Messages.CollapseOld" },
    {   N_("Automatically lock at warp wells"),  1, "", BoolEditor,       "Chart.Scanner.WarpWells" },
    {   N_("Taxation predicts relative growth"), 1, "", BoolEditor,       "Tax.PredictRelative" },
    {   N_("Instant battle result"),             1, "", BoolEditor,       "VCR.InstantResult" },
    //   Guess ship positions
    //   Starcharts file (CHARTx.CC)
    //   Auto load...
    //   If UTILx.DAT available...
    //   Clear console on exit
    //   Reset info dialogs!
    // Hardware Options
    //   Use linear framebuffer
    //   Use VESA Palette setting
    //   Use virtual double-buffering
    // Universe
    //   Delete far planets
    //   Starcharts geometry...
    { N_("Backups"),                             0, "", NoEditor,         "" },
    {   N_("Result file backups..."),            1, "", BackupEditor,     "Backup.Result,%d/backups/result%p.%t" },
    {   N_("UTIL.DAT file backups..."),          1, "", BackupEditor,     "Backup.Util,%d/backups/util%p.%t" },
    {   N_("Turn file backups..."),              1, "", BackupEditor,     "Backup.Turn,%d/backups/turn%p.%t" },
    {   N_("Starchart database backups..."),     1, "", BackupEditor,     "Backup.Chart,%d/backups/chart%p.%t" },
    {   N_("Scripts/Auto Task backups..."),      1, "", BackupEditor,     "Backup.Script,%d/backups/script%p.%t" },
};

static void
addOptionByName(WConfigStorageControl& ctl, Config& cfg, string_t name)
{
    if (ConfigOption* opt = cfg.getConfigOption(name)) {
        ctl.addOption(*opt);
    }
}



WConfigBackupEditor::WConfigBackupEditor(Config& cfg)
    : cfg(cfg)
{ }

string_t
WConfigBackupEditor::toString(string_t arg)
{
    string_t name, deflt;
    strSplit(arg, name, deflt, ",");
    ConfigStringOption* opt = dynamic_cast<ConfigStringOption*>(cfg.getConfigOption(name));
    if (opt == 0) {
        return string_t();
    } else {
        string_t value = (*opt)();
        if (value.empty()) {
            return _("disabled (no backup)");
        } else if (value == deflt) {
            return _("enabled (standard file name)");
        } else {
            return format(_("custom: %s"), value);
        }
    }
}

void
WConfigBackupEditor::edit(string_t arg)
{
    string_t name, deflt;
    strSplit(arg, name, deflt, ",");
    ConfigStringOption* opt = dynamic_cast<ConfigStringOption*>(cfg.getConfigOption(name));
    if (opt) {
        WConfigBackupEditorWindow win(*opt, deflt);
        ui_root->centerWidget(win);
        ui_root->runModal(win);
    }
}


WConfigBackupEditorWindow::WConfigBackupEditorWindow(ConfigStringOption& opt, string_t def)
    : UIWindow(_("Backup Setting"), BLUE_WINDOW, COLOR_GRAY),
      opt(opt),
      def(def),
      h(),
      selection(0),
      input(0)
{
    init();
}

void
WConfigBackupEditorWindow::init()
{
    // VBox
    //   UIRadioButton "disabled"
    //   UIRadioButton "enabled"
    //   UIRadioButton "custom"
    //   HBox
    //     spacer
    //     UIInputLine
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer

    // Input line
    input = &h.add(new UIInputLine(1000, 20, 0, UIInputLine::ilf_Framed, 2003));
    input->setText(opt());

    // Pick default value
    if (opt().empty()) {
        selection.set(0);
        input->setText(def);
    } else if (opt() == def) {
        selection.set(1);
    } else {
        selection.set(2);
    }

    // Build window
    setLayout(UIVBoxLayout::instance5);
    add(h.add(new UIRadioButton(2000, _("disabled (no backup)"), 'd', selection, 0)));
    add(h.add(new UIRadioButton(2001, _("enabled (standard file name)"),  'e', selection, 1)));
    add(h.add(new UIRadioButton(2002, _("custom file name"),   'c', selection, 2)));

    UIGroup& group4 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group4.add(h.add(new UIStaticText("     ", UIStaticText::StaticTitle)));
    group4.add(*input);
    add(group4);

    UIGroup& group5 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group5.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WConfigBackupEditorWindow::onOK));
    group5.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    group5.add(h.add(new UISpacer()));
    group5.add(h.add(new UIButton(_("Help"), 'h')));
    add(group5);

    add(h.add(new UIFocusIterator(2000, 2003, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Vert)));

    selection.sig_changed.add(this, &WConfigBackupEditorWindow::onMove);
    getWidgetById(2000 + selection.get())->setFocusedWidget();
    add(h.add(new WHelpWidget("pcc2:settings:backup")));
    add(h.add(new UIQuit(0)));

    pack();
    onMove();
}

void
WConfigBackupEditorWindow::onMove()
{
    setStateOf(2003, st_Disabled, selection.get() != 2);
}

void
WConfigBackupEditorWindow::onOK()
{
    switch (selection.get()) {
     case 0:
        opt.setAndMarkUpdated(string_t(), ConfigOption::srcUser);
        break;
     case 1:
        opt.setAndMarkUpdated(def, ConfigOption::srcUser);
        break;
     case 2:
        opt.setAndMarkUpdated(input->getText(), ConfigOption::srcUser);
        break;
    }
    stop(1);
}

string_t
WConfigStarchartEditor::toString(string_t /*arg*/)
{
    return _("(dialog)");
}

void
WConfigStarchartEditor::edit(string_t /*arg*/)
{
    doChartConfig();
}


WConfigBooleanEditor::WConfigBooleanEditor(Config& cfg)
    : cfg(cfg)
{ }

string_t
WConfigBooleanEditor::toString(string_t arg)
{
    if (ConfigOption* opt = cfg.getConfigOption(arg)) {
        // This is used for options that have a different converter than Yes/No,
        // so we cannot just convert to Yes/No by hand.
        // But we can translate Yes/No if we get it.
        string_t result = opt->toString();
        if (result == "Yes" || result == "No") {
            result = _(result.c_str());
        }
        return result;
    } else {
        return string_t();
    }
}

void
WConfigBooleanEditor::edit(string_t arg)
{
    ConfigOption* opt = cfg.getConfigOption(arg);
    if (ConfigIntOption* io = dynamic_cast<ConfigIntOption*>(opt)) {
        io->set(!(*io)());
        io->markUpdated(ConfigOption::srcUser);
    } else if (ConfigIntRefOption* iro = dynamic_cast<ConfigIntRefOption*>(opt)) {
        iro->set(!(*iro)());
        iro->markUpdated(ConfigOption::srcUser);
    }
}


string_t
WConfigLabelEditor::toString(string_t /*arg*/)
{
    return _("(dialog)");
}

void
WConfigLabelEditor::edit(string_t /*arg*/)
{
    editLabelExpressions();
}


WConfigGameEditor::WConfigGameEditor(GameOption option)
    : option(option)
{ }

string_t
WConfigGameEditor::toString(string_t /*arg*/)
{
    if (haveDisplayedTurn() && getPlayerId() != 0 && getRealPlayerId() != 0) {
        GGameTurn& current = getCurrentTurn();
        GCommandContainer& cmds = current.getCommands(getPlayerId());
        switch (option) {
         case PasswordOption:
            return current.getGen(getRealPlayerId()).hasPassword()
                ? _("(set)")
                : _("(not set)");
         case RacenameOption:
            return (cmds.getCommand(GCommand::phc_SetRaceName, GCommand::phcrn_Long) != 0
                    || cmds.getCommand(GCommand::phc_SetRaceName, GCommand::phcrn_Short) != 0
                    || cmds.getCommand(GCommand::phc_SetRaceName, GCommand::phcrn_Adjective) != 0)
                ? _("(changed)")
                : _("(not changed)");
         case RemoteOption:
            if (const GCommand* cmd = cmds.getCommand(GCommand::phc_RemoteDefault, 0)) {
                const string_t& arg = cmd->getArg();
                if (arg.size() > 0 && std::tolower(arg[0]) == 'f') {
                    return _("(forbidden by default)");
                } else {
                    return _("(allowed by default)");
                }
            } else {
                return _("(not changed)");
            }
         case FilterOption:
            if (const GCommand* cmd = cmds.getCommand(GCommand::phc_Filter, 0)) {
                if (ValueBoolParser::instance.parse(cmd->getArg())) {
                    return _("(filter enabled)");
                } else {
                    return _("(filter disabled)");
                }
            } else {
                return _("(not changed)");
            }
         case LanguageOption:
            if (const GCommand* cmd = cmds.getCommand(GCommand::phc_Language, 0)) {
                return cmd->getArg();
            } else {
                return _("(not changed)");
            }
         case GetConfigOption:
            return cmds.getCommand(GCommand::phc_SendConfig, 0) != 0
                ? _("(requested)")
                : _("(not requested)");
         case GetFCodeOption:
            return cmds.getCommand(GCommand::phc_SendFcodes, 0) != 0
                ? _("(requested)")
                : _("(not requested)");
         case GetRaceNameOption:
            return cmds.getCommand(GCommand::phc_SendRacenames, 0) != 0
                ? _("(requested)")
                : _("(not requested)");
        }
    }
    return string_t();
}

void
WConfigGameEditor::edit(string_t /*arg*/)
{
    if (haveDisplayedTurn() && getPlayerId() != 0 && getRealPlayerId() != 0) {
        GGameTurn& current = getCurrentTurn();
        GCommandContainer& cmds = current.getCommands(getPlayerId());
        switch (option) {
         case PasswordOption:
            doPasswordChange(current.getGen(getPlayerId()));
            break;
         case RacenameOption:
            editRaceNames(cmds);
            break;
         case RemoteOption:
            editRemote(cmds);
            break;
         case FilterOption:
            editFilter(cmds);
            break;
         case LanguageOption:
            editLanguage(cmds);
            break;
         case GetConfigOption:
            toggleCommand(cmds, GCommand::phc_SendConfig, 0);
            break;
         case GetFCodeOption:
            toggleCommand(cmds, GCommand::phc_SendFcodes, 0);
            break;
         case GetRaceNameOption:
            toggleCommand(cmds, GCommand::phc_SendRacenames, 0);
            break;
        }
    }
}

void
WConfigGameEditor::editRemote(GCommandContainer& cmds)
{
    enum { Yes, No, Cancel };
    UIMessageBox box(RichParser::textFromXml(_("Do you want newly-built ships to be available for remote control "
                                               "by allies by default?\n\n"
                                               "<small>You can give this permission for each ship individually "
                                               "using <kbd>R</kbd> on the ship screen. The recommended setting "
                                               "is <b>No</b>.</small>")),
                     _("Set remote control defaults..."));
    UIKeyString y(_("Yes")), n(_("No"));
    box.addButton(Yes, y.getString(), y.getKey());
    box.addButton(No,  n.getString(), n.getKey());
    box.addButton(Cancel, _("Cancel"), SDLK_ESCAPE);
    switch (box.run()) {
     case Yes: cmds.setCommand(GCommand::phc_RemoteDefault, 0, "allow"); break;
     case No:  cmds.setCommand(GCommand::phc_RemoteDefault, 0, "forbid"); break;
     case Cancel: break;
    }
}

void
WConfigGameEditor::editFilter(GCommandContainer& cmds)
{
    enum { Yes, No, Cancel };
    UIMessageBox box(RichParser::textFromXml(_("Do you want the host to filter out most bulk messages?\n\n"
                                               "<small>When the filter is enabled, messages like sensor sweep "
                                               "are removed from the result file, reducing its size. "
                                               "PCC2 can still display the information because it is also "
                                               "received in the <tt>util.dat</tt> file. "
                                               "The recommended setting is <b>No</b> to keep the messages. "
                                               "You can then use PCC2's built-in filter to determine with "
                                               "finer granularity what you want to see.</small>")),
                     _("Set message detail..."));
    UIKeyString y(_("Yes")), n(_("No"));
    box.addButton(Yes, y.getString(), y.getKey());
    box.addButton(No,  n.getString(), n.getKey());
    box.addButton(Cancel, _("Cancel"), SDLK_ESCAPE);
    switch (box.run()) {
     case Yes: cmds.setCommand(GCommand::phc_Filter, 0, "yes"); break;
     case No:  cmds.setCommand(GCommand::phc_Filter, 0, "no"); break;
     case Cancel: break;
    }
}

void
WConfigGameEditor::editLanguage(GCommandContainer& cmds)
{
    // Build list
    static const char*const languages[] = {
        "English",
        "German",
        "French",
        "Spanish",
        "Italian",
        "Dutch",
        "Russian",
        "Estonian",
        "Polish",
        "NewEnglish",
    };
    StringList list;
    for (size_t i = 0; i < countof(languages); ++i) {
        list.add(i, languages[i]);
    }
    list.add(-1, _("(no change)"));

    // Prepare
    int current = -1;
    if (const GCommand* cmd = cmds.getCommand(GCommand::phc_Language, 0)) {
        for (size_t i = 0; i < countof(languages); ++i) {
            if (strCaseCmp(languages[i], cmd->getArg()) == 0) {
                current = i;
                break;
            }
        }
    }

    // Do it
    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    box.scrollToKey(current);
    if (box.doStandardListbox(_("Set message language..."), "pcc2:settings")) {
        int key = box.getKey();
        if (key < 0 || key >= int(countof(languages))) {
            cmds.removeCommand(GCommand::phc_Language, 0);
        } else {
            cmds.setCommand(GCommand::phc_Language, 0, languages[key]);
        }
    }
}

void
WConfigGameEditor::toggleCommand(GCommandContainer& cmds, GCommand::Type type, int id)
{
    if (!cmds.removeCommand(type, id)) {
        cmds.setCommand(type, id, string_t());
    }
}

WPreferencesDialog::WPreferencesDialog(Config& cfg)
    : UIWindow(_("Settings"), BLUE_WINDOW, COLOR_GRAY),
      cfg(cfg),
      settings(0),
      game_settings(0),
      details(0),
      editor(0),
      game_editor(0),
      storage(0),
      editors(),
      game_editor_obj()
{
    init();
}

void
WPreferencesDialog::init()
{
    // VBox
    //   UICardGroupTab
    //   UICardGroup "card_group2"
    //     HBox "group21"
    //       UICardGroup "card_group211"
    //         UITreeWidget "Options"
    //         WConfigView "Details"
    //       VBox "212"
    //         WConfigEditor
    //         UISpacer            / later: doc
    //         WConfigStorageControl
    //     HBox "group22"
    //       UITreeWidget "game_settings"
    //       VBox "group221"
    //         WConfigEditor "game_editor"
    //         UISpacer
    //   HBox "group3"
    //     UIButton "OK"
    //     UISpacer
    //     UIButton "Help"

    // Set up editors
    editors.push_back(0);
    editors.push_back(new WConfigOptionEditor(cfg));
    editors.push_back(new WConfigBackupEditor(cfg));
    editors.push_back(new WConfigStarchartEditor());
    editors.push_back(new WConfigBooleanEditor(cfg));
    editors.push_back(new WConfigLabelEditor());

    // Create these first so they die last, and card_group's sig_focus_changed
    // does not get in our way during destruction.
    settings = &h.add(new UITreeWidget(1001, 20, 25*font_ems[FONT_NORMAL]));
    details = &h.add(new WConfigView(1002, 20, 25*font_ems[FONT_NORMAL]));
    game_settings = &h.add(new UITreeWidget(1000, 20, 25*font_ems[FONT_NORMAL]));

    UICardGroup&    card_group2   = h.add(new UICardGroup());
    UICardGroup&    card_group211 = h.add(new UICardGroup());
    UICardGroupTab& card_tabs     = h.add(new UICardGroupTab(card_group2));
    card_tabs.observe(card_group211.sig_focus_changed);

    UIGroup& group21  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group212 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group22  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group222 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group3   = h.add(new UIGroup(UIHBoxLayout::instance5));

    setLayout(UIVBoxLayout::instance5);
    if (haveDisplayedTurn()) {
        card_tabs.addTab(1000, UIKeyString(_("Game Settings")));
    }
    card_tabs.addTab(1001, UIKeyString(_("Options")));
    card_tabs.addTab(1002, UIKeyString(_("Configuration File")));
    add(card_tabs);
    for (uint_t i = 0; i < countof(config_tree); ++i) {
        settings->addItem(i, config_tree[i].level, _(config_tree[i].name));
    }
    settings->updateAfterModification();
    settings->sig_moved.add(this, &WPreferencesDialog::onMove);

    if (haveDisplayedTurn()) {
        // Build the list.
        // The list is never empty:
        // - THost always has the "Password" option
        // - PHost always has the "Filter" option
        if (!host.isPHost() || !config.DisablePasswords()) {
            game_settings->addItem(PasswordOption, 0, _(game_option_names[PasswordOption]));
        }
        if (host.isPHost() && config.CPEnableRaceName()) {
            game_settings->addItem(RacenameOption, 0, _(game_option_names[RacenameOption]));
        }
        if (host.isPHost() && config.CPEnableRemote()) {
            game_settings->addItem(RemoteOption, 0, _(game_option_names[RemoteOption]));
        }
        if (host.isPHost()) {
            game_settings->addItem(FilterOption, 0, _(game_option_names[FilterOption]));
        }
        if (host.isPHost() && config.CPEnableLanguage()) {
            game_settings->addItem(LanguageOption, 0, _(game_option_names[LanguageOption]));
        }
        if (host.isPHost() && config.CPEnableSend()) {
            game_settings->addItem(-1, 0, _("Request file from host"));
            game_settings->addItem(GetConfigOption,   1, _(game_option_names[GetConfigOption]));
            game_settings->addItem(GetFCodeOption,    1, _(game_option_names[GetFCodeOption]));
            game_settings->addItem(GetRaceNameOption, 1, _(game_option_names[GetRaceNameOption]));
        }
        game_settings->updateAfterModification();
        game_settings->sig_moved.add(this, &WPreferencesDialog::onMove);
    }

    card_group211.add(*settings);
    details->setNameColumnWidth(13);
    details->setContent(cfg);
    details->setHighlightSource(ConfigOption::srcUser);
    details->sig_moved.add(this, &WPreferencesDialog::onMove);
    card_group211.add(*details);
    card_group211.sig_focus_changed.add(this, &WPreferencesDialog::onMove);
    group21.add(UIFrameGroup::wrap(h, card_group211));

    editor = &h.add(new WConfigEditor());
    group212.add(*editor);
    group212.add(h.add(new UISpacer()));
    storage = &h.add(new WConfigStorageControl(true));
    group212.add(*storage);
    group21.add(group212);

    game_editor = &h.add(new WConfigEditor());
    group222.add(*game_editor);
    group222.add(h.add(new UISpacer()));
    group22.add(UIFrameGroup::wrap(h, *game_settings));
    group22.add(group222);
    card_group2.add(group21);
    card_group2.add(group22);
    card_group2.sig_focus_changed.add(this, &WPreferencesDialog::onMove);
    add(card_group2);

    group3.add(h.add(new UIButton(1, _("Close"), SDLK_ESCAPE)).addStop());
    group3.add(h.add(new UISpacer()));
    group3.add(h.add(new UIButton(_("Help"), 'h')));
    add(group3);
    add(h.add(new WHelpWidget("pcc2:settings")));
    add(h.add(new UIQuit(1)));

    pack();

    details->observe(editor->sig_change);
    details->observe(storage->sig_change);
    if (haveDisplayedTurn()) {
        game_settings->setFocusedWidget();
    } else {
        settings->setFocusedWidget();
    }
    editor->sig_change.add(this, &WPreferencesDialog::onMove);
    game_editor->sig_change.add(this, &WPreferencesDialog::onMove);

    onMove();
}

void
WPreferencesDialog::onMove()
{
    if (game_settings->hasState(st_Focused)) {
        int id = game_settings->getItemId(game_settings->getCurrentItemIndex());
        if (id >= 0) {
            game_editor_obj.reset(new WConfigGameEditor(GameOption(id)));
            game_editor->set(_(game_option_names[id]), game_editor_obj.get(), string_t());
        } else {
            game_editor->set(string_t(), 0, string_t());
        }
    } else if (settings->hasState(st_Focused)) {
        storage->clear();

        const ConfigItem& item = config_tree[settings->getItemId(settings->getCurrentItemIndex())];
        editor->set(_(item.name), editors[item.editor], item.arg);
        switch (item.editor) {
         case NoEditor:
            break;
         case DefaultEditor:
         case BoolEditor:
            addOptionByName(*storage, cfg, item.arg);
            break;
         case BackupEditor:
            addOptionByName(*storage, cfg, strFirst(item.arg, ","));
            break;
         case StarchartEditor:
            for (int i = 0, n = getNumChartOptions(); i < n; ++i) {
                storage->addOption(getChartOption(i));
            }
            for (int i = 0; i < ConfigMarkerOption::NUM_CANNED_MARKERS; ++i) {
                storage->addOption(ConfigMarkerOption::getCannedMarker(i));
            }
            for (int i = 0; i < lm_MAX; ++i) {
                storage->addOption(getLockModeConfigOption(TLockMode(i)));
            }
            break;
         case LabelEditor:
            addOptionByName(*storage, cfg, "Label.Ship");
            addOptionByName(*storage, cfg, "Label.Planet");
            break;
        }
    } else if (details->hasState(st_Focused)) {
        storage->clear();
        if (ConfigOption* opt = details->getCurrentOption()) {
            storage->addOption(*opt);
            editor->set(opt->getName(), editors[DefaultEditor], opt->getName());
        } else {
            editor->set(string_t(), 0, string_t());
        }
    }
}

/** Display the preferences dialog. */
void
doPreferencesDialog()
{
    WPreferencesDialog dlg(getUserPreferences());
    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}
