/**
  *  \file client/tiles/baseorder.cc
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/baseorder.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/spec.h"
#include "game/planet.h"
#include "game/tables.h"
#include "game/ship.h"
#include "ui/table.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "game/univ.h"

/***************************** WBaseOrderTile ****************************/

WBaseOrderTile::WBaseOrderTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, 100), selection)
{
    setLayout(LeftJustified, 5, true);
    addControl(GfxRect(0,  2, 20, 20), h.add(new UIButton("B", 'b')));
    addControl(GfxRect(0, 27, 20, 20), h.add(new UIButton(cm_Base_Recycle, "R", 'r')).addCommand());
    addControl(GfxRect(0, 52, 20, 20), h.add(new UIButton("M", 'm')));
    addControl(GfxRect(0, 77, 20, 20), h.add(new UIButton("A", 'a')));
}

void
WBaseOrderTile::drawData(GfxCanvas& can)
{
    GPlanet* p;
    if ((p = dynamic_cast<GPlanet*>(getCurrentObject())) != 0 && p->hasFullBaseData()) {
        /* headers */
        GfxContext ctx;
        ctx.useColorScheme(getSkin());
        ctx.useCanvas(can);
        ctx.useFont(*fonts[FONT_NORMAL]);
        ctx.setColor(UIColor::tc_Static);
        ctx.setSolidBackground();
        outText(ctx, 35, 55, _("Mission:"));
        outText(ctx, 35, 80, _("Build Fighters/Torpedoes"));

        /* ship */
        GUniverse* univ = getCurrentUniverse();
        int hull = p->getBaseBuildHull();
        int clnid = p->findShipCloningHere(*univ);
        string_t text;

        if (hull) {
            if (clnid) {
                ctx.setColor(UIColor::tc_Red);
                text = format(_("Cloning, and building %s"),
                              getHull(hull).getName());
            } else {
                ctx.setColor(UIColor::tc_Yellow);
                text = format(_("Building a %s"),
                              getHull(hull).getName());
            }
        } else {
            if (clnid) {
                ctx.setColor(UIColor::tc_Yellow);
                text = format(_("Cloning %s"), univ->getShip(clnid).getName(GObject::PlainName));
            } else {
                ctx.setColor(UIColor::tc_Static);
                text = _("Build a new ship");
            }
        }

        /* position */
        mp16_t qpos = p->getBaseQueuePosition();
        int w = extent.w;
        if (hull != 0 && qpos.isKnown()) {
            string_t qtext(format(" Q:%d", int(qpos)));
            int qw = fonts[FONT_NORMAL]->getTextWidth(qtext);
            w -= qw;
            outTextF(ctx, w, 5, qw, qtext);
        }

        outTextF(ctx, 35, 5, w-35, text);

        /* fix/recycle */
        if (p->getBaseShipyardOrder()) {
            ctx.setColor(UIColor::tc_Yellow);
            outTextF(ctx, 35, 30, extent.w-35,
                     format(p->getBaseShipyardOrder() == 1 /* FIXME: we have an enum for that */
                            ? _("Repairing %s")
                            : _("Recycling %s"),
                            univ->getShip(p->getBaseShipyardId()).getName(GObject::LongName)));
        } else {
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, 35, 30, extent.w-35, _("Repair/recycle a starship"));
        }

        /* mission */
        const int mx = 65 + fonts[FONT_NORMAL]->getTextWidth(_("Mission:"));
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, mx, 55, extent.w-mx, getBaseMissionName(p->getBaseMission()));
    }
}



WNarrowBaseOrderTile::WNarrowBaseOrderTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(250, 4*font_heights[FONT_NORMAL]), selection)
{ }

void
WNarrowBaseOrderTile::drawData(GfxCanvas& can)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getCurrentObject());
    if (!p || !p->hasFullBaseData())
        return;

    /* Building foo
       Repairing foo
       Mission:  repair
       FCode:    xxx */

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    GfxRect in = getInnerSize(0, 0);
    int x = in.x;
    int y = in.y;
    int w = in.w;

    /* Build */
    GUniverse* univ = getCurrentUniverse();
    int hull = p->getBaseBuildHull();
    int clnid = p->findShipCloningHere(*univ);
    string_t text;

    if (hull) {
        if (clnid) {
            ctx.setColor(UIColor::tc_Red);
            text = format(_("Cloning, and building %s"),
                          getHull(hull).getName());
        } else {
            ctx.setColor(UIColor::tc_Yellow);
            text = format(_("Building a %s"),
                          getHull(hull).getName());
        }
    } else {
        if (clnid) {
            ctx.setColor(UIColor::tc_Yellow);
            text = format(_("Cloning %s"), univ->getShip(clnid).getName(GObject::PlainName));
        } else {
            ctx.setColor(UIColor::tc_Static);
            text = _("Not building a ship");
        }
    }
    
    outTextF(ctx, x, y, w, text);
    y += font_heights[FONT_NORMAL];

    /* Shipyard */
    if (p->getBaseShipyardOrder()) {
        ctx.setColor(UIColor::tc_Yellow);
        outTextF(ctx, x, y, w, 
                 format(p->getBaseShipyardOrder() == 1 /* FIXME: we have an enum for that */
                        ? _("Repairing %s")
                        : _("Recycling %s"),
                        univ->getShip(p->getBaseShipyardId()).getName(GObject::LongName)));
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, _("Not working on a starship"));
    }
    y += font_heights[FONT_NORMAL];

    /* Mission/FCode */
    UITableWriter tw(ctx, GfxRect(x, y, w, in.getBottomY() - y), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Mission:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Friendly Code:"));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, getBaseMissionName(p->getBaseMission()));
    tw.writeVariableCell(UIColor::tc_Green, p->getFCode());

    tw.finish();
}
