/**
  *  \file flak/dlg-select.cc
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "flak/dlg-select.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "ui/listbox.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "game/spec.h"
#include "ui/commands.h"
#include "game/racename.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/framegroup.h"
#include "util/recode.h"
#include "ui/widgets/button.h"
#include "util/unicodechars.h"
#include "client/widgets/vcrinfomain.h"
#include "client/widgets/vcrinfohead.h"
#include "client/chart.h"
#include "client/dialogs/combatdlg.h"

namespace {
    class FlakParticipantList : public UIListbox {
        const FlakVcrEntry& entry;
        std::vector<int> content;
        enum {
            flShip   =  8192,
            flFleet  = 16384,
            flIdMask =  2047
        };
        void onMove(int);
     public:
        FlakParticipantList(FlakVcrEntry& entry);

        bool isAccessible(int pos);
        void drawPart(GfxCanvas& can, int from, int to);
        void getLayoutInfo(LayoutInfo& info);

        Signal1<int> sig_moved_to_ship;
    };

    class WFlakVcrInfoWindow : public UIWindow {
        ObjectHolder h;
        FlakParticipantList* fpl;
        WVcrInfoMain* main;
        WVcrInfoHeader* head;
        FlakVcrEntry& battle;
     public:
        WFlakVcrInfoWindow(FlakVcrEntry& b)
            : UIWindow(_("Combat Details"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
              h(),
              fpl(0),
              main(0),
              head(0),
              battle(b)
            { init(); }
        bool handleEvent(const UIEvent& event, bool second_pass);
     private:
        void init();
        void onMove(int i);
    };
}

FlakParticipantList::FlakParticipantList(FlakVcrEntry& entry)
    : UIListbox(0, 10, 20 * font_ems[FONT_NORMAL], 0), entry(entry)
{
    const FlakBattle& b = entry.getBattle();
    for (int i = 0; i < b.getNumFleets(); ++i) {
        const FlakFleet& fleet = b.getFleetByNumber(i);
        content.push_back(i + flFleet);
        for (int s = 0; s < fleet.data.num_ships; ++s)
            content.push_back(s + fleet.data.first_ship + flShip);
    }
    setSize(content.size());
    scrollTo(0, 1);
    sig_moved.add(this, &FlakParticipantList::onMove);
}

void
FlakParticipantList::onMove(int)
{
    if (isAccessible(getCurrent()))
        sig_moved_to_ship.raise(content[getCurrent()] & flIdMask);
}

bool
FlakParticipantList::isAccessible(int pos)
{
    return (pos >= 0) && (pos < content.size()) && ((content[pos] & flShip) != 0);
}

void
FlakParticipantList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    const FlakBattle& b = entry.getBattle();
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        int ele = getTop() + i;
        if (ele < content.size()) {
            int id = content[ele] & flIdMask;
            if (content[ele] & flShip) {
                ctx.useFont(*fonts[FONT_NORMAL]);
                outTextF(ctx, area.x + 10, area.y, area.w - 10, convertGameToUtf8(getBasicString(b.getShipByNumber(id).data.name)));
            } else {
                ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
                ctx.setColor(getPlayerColor(b.getFleetByNumber(id).data.player));
                outTextF(ctx, area.x, area.y, area.w,
                         format(_("Fleet %d (%s)"), id,
                                player_racenames.getAdjName(b.getFleetByNumber(id).data.player)));
            }
        }
    }
}

void
FlakParticipantList::getLayoutInfo(LayoutInfo& info)
{
    info.flags     = li_Flexible;
    info.pref_size = font_sizes[FONT_NORMAL].scaledBy(10, 20);
    info.min_size  = font_sizes[FONT_NORMAL].scaledBy(3, 5);
}

/*************************** WFlakVcrInfoWindow **************************/

bool
WFlakVcrInfoWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_INSERT:
            /* Add to sim, after fight */
            main->addToSim(true);
            return true;

         case SDLK_INSERT + ss_Shift:
            /* Add to sim, before fight */
            main->addToSim(false);
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WFlakVcrInfoWindow::init()
{
    // VBox
    //   HBox [content]
    //     UIFrameGroup [listFrame]
    //       FlakParticipantList
    //     VBox [detail]
    //       WVcrInfoHeader
    //       WVcrInfoMain
    //       UISpacer
    //   HBox [buttons]
    //     UISpacer
    //     UIButton "Close"
    static const UIKeyTranslator::Item xl[] = {
        SDLK_ESCAPE, cm_Escape,
        ' ', cm_Escape
    };

    UIGroup& content        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& buttons        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& detail         = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIFrameGroup& listFrame = h.add(new UIFrameGroup(UIHBoxLayout::instance0));

    fpl  = &h.add(new FlakParticipantList(battle));
    main = &h.add(new WVcrInfoMain(true));
    head = &h.add(new WVcrInfoHeader(battle, 0));
    fpl->sig_moved_to_ship.add(this, &WFlakVcrInfoWindow::onMove);
    head->getList().setState(st_Disabled, true);

    setLayout(UIVBoxLayout::instance5);
    add(content);
    add(buttons);

    buttons.add(h.add(new UIButton(cm_Escape, _("Close"), SDLK_RETURN)).addStop());
    buttons.add(h.add(new UIButton(_("Ins - Sim"), SDLK_INSERT)));
    buttons.add(h.add(new UISpacer()));
    // buttons.add(h.add(new UIButton("H", 'h')));
    buttons.add(h.add(new UIKeyTranslator(xl, countof(xl))));

    listFrame.add(*fpl);
    content.add(listFrame);
    content.add(detail);
    detail.add(*head);
    detail.add(*main);
    detail.add(h.add(new UISpacer()));
    pack();
    onMove(0);    /* FIXME: what if FlakParticipantList starts on another item? */
}

void
WFlakVcrInfoWindow::onMove(int i)
{
    main->set(battle, i, battle.getObject(i, false).getHull());
    head->getList().setVcrEntry(battle, i);
}

/**************************** FlakVcrSelector ****************************/

FlakVcrSelector::FlakVcrSelector(GFlakVcrDatabase& db)
    : db(db), current(0),
      btn_l(0, "L", 'l'),
      btn_tab(0, "Tab", SDLK_TAB)
{
    add(btn_l, GfxRect(0, 0, 25, 25));
    add(btn_tab, GfxRect(0, 0, 40, 25));
    btn_tab.setState(st_Disabled, db.getNumBattles() <= 1);
    btn_l.sig_fired.add(this, &FlakVcrSelector::showInfo);
    btn_tab.sig_fired.add(this, &FlakVcrSelector::showDiagram);
}

FlakVcrSelector::~FlakVcrSelector()
{ }

void
FlakVcrSelector::drawContent(GfxCanvas& can)
{
    FlakVcrEntry& b = db.getBattle(current);

    int y = getExtent().y;
    int x = getExtent().x;
    int w = getExtent().w;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_TITLE]);
    drawBackground(ctx, getExtent());

    outTextF(ctx, x, y, w, format(_("Battle %d of %d:"), current+1, db.getNumBattles()));
    ctx.setTextAlign(2, 0);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, x+w, y, w, _("FLAK"));
    ctx.setTextAlign(0, 0);
    y += font_heights[FONT_TITLE];
    y += 5;

    int nships  = b.getBattle().getNumShips();
    int nfleets = b.getBattle().getNumFleets();
    outTextF(ctx, x, y, w, format(_("Location: (%d,%d)"), b.getPositionX(), b.getPositionY()));
    y += font_heights[FONT_NORMAL];
    outTextF(ctx, x, y, w, format(_("Time: %d:%02d"), b.getTotalTime() / 60, b.getTotalTime() % 60));
    y += font_heights[FONT_NORMAL];
    outTextF(ctx, x, y, w, format(_("%d unit%!1{s%} in %d group%!1{s%}:"), nships, nfleets));
    y += font_heights[FONT_NORMAL];

    int n = (getExtent().h - (y - getExtent().y)) / font_heights[FONT_NORMAL];
    if (nships <= n) {
        // Enough room for ships
        for (int i = 0; i < nships; ++i) {
            const GVcrObject& sh = b.getObject(i, false);
            ctx.setColor(getPlayerColor(sh.getOwner()));
            outTextF(ctx, x+15, y, w-15,
                     sh.isPlanet()
                     ? format(_("%s (#%d, a %s planet)")) << sh.getName() << sh.getId() << player_racenames.getAdjName(sh.getOwner())
                     : sh.getHull() != 0
                     ? format(_("%s (#%d, a %s %s)"))     << sh.getName() << sh.getId() << player_racenames.getAdjName(sh.getOwner()) << getHull(sh.getHull()).getShortName()
                     : format(_("%s (#%d, a %s ship)"))   << sh.getName() << sh.getId() << player_racenames.getAdjName(sh.getOwner()));
            y += font_heights[FONT_NORMAL];
        }
    } else if (nfleets <= n) {
        // Enough room for fleets
        for (int i = 0; i < nfleets; ++i) {
            const FlakFleet& f = b.getBattle().getFleetByNumber(i);
            const GVcrObject& o = b.getObject(f.data.first_ship, false);
            string_t text;
            ctx.setColor(getPlayerColor(f.data.player));
            if (f.data.num_ships == 1 && o.isPlanet()) {
                text = format(_("a %s planet"), player_racenames.getAdjName(f.data.player));
            } else {
                text = format(_("%d %s unit%0$!d%!1{s%}"), f.data.num_ships, player_racenames.getAdjName(f.data.player));
            }
            outTextF(ctx, x+15, y, w-15, text);
            y += font_heights[FONT_NORMAL];
        }
    } else {
        // Race summary
        GPlayerArray<int> sum;
        sum.setAll(0);
        for (int i = 0; i < nfleets; ++i) {
            ++sum[b.getBattle().getFleetByNumber(i).data.player];
        }
        for (int i = 1; i <= NUM_OWNERS; ++i) {
            if (sum[i] != 0) {
                if (n > 0) {
                    --n;
                    ctx.setColor(getPlayerColor(i));
                    outTextF(ctx, x+15, y, w-15, format(_("%d %s group%0$!d%!1{s%}"), sum[i], player_racenames.getAdjName(i)));
                    y += font_heights[FONT_NORMAL];
                }
            }
        }
    }
    can.update(getExtent());
}

void
FlakVcrSelector::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(30, 13);
}

void
FlakVcrSelector::next(int)
{
    setCurrent(getCurrent()+1);
}

void
FlakVcrSelector::previous(int)
{
    setCurrent(getCurrent()-1);
}

void
FlakVcrSelector::showInfo()
{
    WFlakVcrInfoWindow win(db.getBattle(current));
    ui_root->centerWidget(win);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);
}

void
FlakVcrSelector::showDiagram()
{
    int n = doCombatDialog(db);
    if (n >= 0) {
        setCurrent(n);
    }
}

void
FlakVcrSelector::setCurrent(int n)
{
    if (n >= db.getNumBattles())
        n = db.getNumBattles()-1;
    if (n < 0)
        n = 0;

    if (n != current) {
        current = n;
        sig_scrolled.raise();
        drawWidget();
    }
}

void
FlakVcrSelector::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    const GfxRect& x = getExtent();
    btn_l.setExtent(GfxRect(x.getRightX() - 25, x.getBottomY() - 55, 25, 25), true);
    btn_tab.setExtent(GfxRect(x.getRightX() - 40, x.getBottomY() - 25, 40, 25), true);
}
