/**
  *  \file int/if/userif.h
  *  \brief Interpreter: User Interface Binding
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_IF_USERIF_H
#define PCC_V2_INT_IF_USERIF_H

#include <memory>
#include "cpluslib/ptrvector.h"
#include "cpluslib/string.h"
#include "int/data.h"

class IntValue;
class IntContext;

/** User interface property identifier. */
enum IntUserInterfaceProperty {
    iuiScreenNumber,               ///< Current screen number (UI.Screen).
    iuiScreenRegistered,           ///< true iff current screen is on screen history.
    iuiSimFlag,                    ///< true iff we are in the simulator (System.Sim).
    iuiScanX,                      ///< Scanner coordinate, X (UI.X). Can be null.
    iuiScanY,                      ///< Scanner coordinate, Y (UI.Y). Can be null.
    iuiChartX,                     ///< Starchart coordinate, X (Chart.X). Can be null.
    iuiChartY,                     ///< Starchart coordinate, Y (Chart.Y). Can be null.
    iuiKeymap                      ///< Keymap. A keymap object, do not expose to users! Can be null.
};

/** User interface command identifier. */
enum IntUserInterfaceCommand {
    iucPopupConsole,               ///< Open console. No parameters.
    iucExitRace,                   ///< Exit to race selection. No parameters.
    iucExitClient,                 ///< Exit to operating system. No parameters.
    iucOverlayMessage,             ///< Show decaying overlay message. Parameter: string.
    iucChartSetView,               ///< Set chart view. Parameter: string.
    iucGoToScreen,                 ///< Open a screen. Parameter: screenId:int. When going to the starchart, this can manipulate the position.
    iucGoToChart,                  ///< Open starchart. No parameter. Does not manipulate the position.
    iucGoToMinefield,              ///< Show a minefield. Shows the dialog.
    iucGoToUfo,                    ///< Show an Ufo. Shows the dialog.
    iucGoToStorm,                  ///< Show an ion storm. Shows the dialog.
    iucUseKeymap,                  ///< Use secondary keymap. Parameter: keymap:Keymap, prefix:Int.
    iucExitDialog                  ///< Exit current dialog. Optional parameter: any object.
};

/** User interface command event.
    If the script side requests something, it posts a command of this type
    (see IntUserInterfaceBinding::postEvent).
    It contains the command code as well as optional parameters. */
struct IntUserInterfaceEvent {
    /** Command to execute. */
    IntUserInterfaceCommand command;

    /** Parameters for this command. */
    IntDataSegment args;

    IntUserInterfaceEvent(IntUserInterfaceCommand cmd)
        : command(cmd),
          args()
        { }
};

/** User interface binding. This is the interface used by scripts. */
class IntUserInterfaceBinding {
 public:
    IntUserInterfaceBinding();
    virtual ~IntUserInterfaceBinding();

    /** Get property.
        \param p Property to read */
    virtual IntValue* getProperty(IntUserInterfaceProperty p) = 0;

    /** Set property.
        \param p Property to modify
        \param v New value */
    virtual void setProperty(IntUserInterfaceProperty p, IntValue* v) = 0;

    /** Post an event.
        This function must:
        - decide whether this command shall be executed
        - remember the event
        - cause its invoking event loop to exit.

        It can assume that when the event loop is reached that this IntUserInterfaceBinding
        will be the topmost one.

        This function can be called from a variety of contexts:
        - from a script context (exceptions thrown will be reflected back to the script)
        - from an event loop (if C++ invokes the binding)
        - from a child's command processing */
    virtual void postEvent(std::auto_ptr<IntUserInterfaceEvent> event) = 0;

    /** Enumerate contexts of this place in user interface.
        Used for creation of new processes from here.
        \param ctx [out] Newly-created contexts are appended here */
    virtual void enumContexts(ptr_vector<IntContext>& ctx) = 0;

    static IntUserInterfaceBinding& get();
};

/** Root binding. An instance of this type always exists.
    This grounds all requests; it is an error for a request to reach this class. */
class IntUserInterfaceRootBinding : public IntUserInterfaceBinding {
 public:
    IntUserInterfaceRootBinding();
    virtual ~IntUserInterfaceRootBinding();

    virtual IntValue* getProperty(IntUserInterfaceProperty p);
    virtual void setProperty(IntUserInterfaceProperty p, IntValue* v);
    virtual void postEvent(std::auto_ptr<IntUserInterfaceEvent> event);
    virtual void enumContexts(ptr_vector<IntContext>& ctx);
};

/** Child binding. All screens inherit from this.
    This manages stackability of bindings. */
class IntUserInterfaceChildBinding : public IntUserInterfaceBinding {
 public:
    IntUserInterfaceChildBinding();
    virtual ~IntUserInterfaceChildBinding();

 protected:
    IntUserInterfaceBinding& parent;
};

#endif
