#!/usr/bin/perl -w
#
#  Convert the PHost/PDK config definitions for use with PCC II
#
#  Whereas this originally replicated PHost's config parser quite
#  accurately, and generated the required structure and code from
#  PHost's definitions directly, we now use a different approach:
#  In the first step, the original definitions are incorporated
#  into an intermediate file. A definition already in that file
#  will stay there. In the second step, we generate the required
#  source snippets from that intermediate file.
#
#  This makes it easy to deviate from the PHost definition of the
#  config file:
#  - to support options which THost has and PHost doesn't
#  - to support obsolete options
#  - to anticipate future arrayisation
#
#  Invocation:
#    perl confgen.pl [--mode=XXX] [files...]
#  It reads and merges all files, and generates output according
#  to mode:
#    xml
#      read .hi files, generate basic XML
#    decl, set, init, set-depend, tim
#      generate code snippet for cut&pasting into config.cc/config.h
#    all
#      generate all code snippets at once
#
#  Recipes: to update with a new config.hi from PHost/PDK, do
#    cp /path/to/config.hi4 phost/
#    perl scripts/confgen.pl > config.new && mv config.new phost/config.xml
#    perl scripts/confgen.pl --mode=all
#  ...and cut&paste the snippets where they belong
#
use strict;

# $option = {
#    typ => ConfigStdOption etc.,
#    name => AllowWebMines etc.,
#    init => "(X, parse_int)",
#    set => method invocation to set default value
# }
my @option_list;
my %options_by_name;

my @hconfig_ranges  = (0, 10, 186, 288, 298, 302, 328, 336, 338, 340);

# This translates input definitions into C++ definitions:
my %translation = (
                Uns16 =>           { "4" =>  { type=>"ConfigIntArrayOption<4>", init=>"(X, ValueIntParser::instance)"  },
                                     "9" =>  { type=>"ConfigIntArrayOption<9>", init=>"(X, ValueIntParser::instance)"  },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueIntParser::instance)",
                                               import=>"importArray16(X, HH, OO)"                       },
                                     "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueIntParser::instance)",
                                               import=>"X.set(getUint16(HH, OO))"                       },
                                     "10" => { type=>"ConfigExpOption",         init=>"(X, ValueIntParser::instance)"  } },
                Uns32 =>           { "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueIntParser::instance)",
                                               import=>"X.set(getUint32(HH, OO))"                       },
                                     "8" =>  { type=>"ConfigIntArrayOption<8>", init=>"(X, ValueIntParser::instance)",
                                               import=>"importArray32(X, HH, OO)"                       },
                                     "2" =>  { type=>"ConfigIntArrayOption<2>", init=>"(X, ValueIntParser::instance)"  },
                                     "10" => { type=>"ConfigExpOption",         init=>"(X, ValueIntParser::instance)"  },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueIntParser::instance)"  } },
                Int16 =>           { "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueIntParser::instance)"  },
                                     "10" => { type=>"ConfigExpOption",         init=>"(X, ValueIntParser::instance)"  },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueIntParser::instance)"  } },
                Int32 =>           { "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueIntParser::instance)"  },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueIntParser::instance)"  } },
                Boolean =>         { "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueBoolParser::instance)",
                                               import=>"X.set(getUint16(HH, OO))"                       },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueBoolParser::instance)" }},
                Tristate =>        { "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueBoolParser::instance)" },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueBoolParser::instance)" }},
                TristateX =>       { "1" =>  { type=>"ConfigIntOption",         init=>"(X, ValueBoolParser::instance)" },
                                     "11" => { type=>"ConfigStdOption",         init=>"(X, ValueBoolParser::instance)" }},
                Language_Def =>    { "*" =>  { type=>"ConfigStringOption",      init=>"(X)" } },
                BuildQueue_Def =>  { "*" =>  { type=>"ConfigStringOption",      init=>"(X)" } },
                ScoreMethod_Def => { "*" =>  { type=>"ConfigStringOption",      init=>"(X)" } },
                char =>            { "*" =>  { type=>"ConfigStringOption",      init=>"(X)" } },
                String =>          { "*" =>  { type=>"ConfigStringOption",      init=>"(X)" } },
                Cost =>            { "11" => { type=>"ConfigCostOption",        init=>"(X)" } },
               );

my $defaultasgn = "";
my $declaration = "";
my $initialisation = "";
#%thost = ();

## Fetch THost configuration
#open TCONF, "<phost/tconfig.hi" or die "can't open tconfig.hi";
#while (<TCONF>) {
#    if (/^TCDefine\((\S+),\s*(\S+),\s*(\S+),\s*(\S+)\)/) {
#	my ($typ, $count, $ofs, $name) = ($1, $2, $3, $4);
#	my $typxl = $typexl{$typ};
#	die "unknown type `$typ'" if !defined $typxl;
#	$thost{$name} = "$typeid{$typxl}, $count, $ofs";
#    } elsif(/^TCDefine/) {
#	die "not matched: $_";
#    }
#}
#close TCONF;

my $mode = 'xml';
if (@ARGV) {
    if ($ARGV[0] =~ /^--mode=(.*)/) {
        shift @ARGV;
        $mode = $1;
    }
}

# PHost configuration
if (!@ARGV) {
    if (-r "phost/config.xml") {
        push @ARGV, "phost/config.xml";
    }
    if (-r "phost/config.hi4") {
        push @ARGV, "phost/config.hi4";
    } else {
        push @ARGV, "phost/config.hi";
    }
    if (-r "phost/tconfig.hi") {
        push @ARGV, "phost/tconfig.hi";
    }
}

if ($mode eq 'xml') {
    processFiles(@ARGV);
    print "<!-- Although this looks like XML, the tool which processes\n";
    print "     it, $0, does not (yet) have a full XML parser, so don't\n";
    print "     be too creative with formatting. -->\n";
    print "<option-set>\n";
    foreach my $o (@option_list) {
        print "  <option name=\"$o->{name}\" type=\"$o->{type}\">\n";
        foreach (sort keys %$o) {
            print "    <$_>$o->{$_}</$_>\n"
              unless $_ eq 'name' || $_ eq 'type';
        }
        print "  </option>\n\n";
    }
    print "</option-set>\n";
} elsif ($mode eq 'decl') {
    processFiles(@ARGV);
    generateDeclaration();
} elsif ($mode eq 'init') {
    processFiles(@ARGV);
    generateInitialisation();
} elsif ($mode eq 'set') {
    processFiles(@ARGV);
    generateSet();
} elsif ($mode eq 'set-depend') {
    processFiles(@ARGV);
    generateSetDepend();
} elsif ($mode eq 'tim') {
    processFiles(@ARGV);
    generateTim();
} elsif ($mode eq 'all') {
    processFiles(@ARGV);
    print "//--- Declaration ---\n";
    generateDeclaration();
    print "\n//--- Initialisation ---\n";
    generateInitialisation();
    print "\n//--- Set ---\n";
    generateSet();
    print "\n//--- Set dependant options ---\n";
    generateSetDepend();
    print "\n//--- Import HConfig ---\n";
    generateTim();
} else {
    print STDERR "$0: invalid mode '$mode'\n";
    exit(1);
}

sub generateDeclaration {
    foreach my $o (@option_list) {
        printf "    %-25s %s;\n", $o->{type}, $o->{name};
    }
}

sub generateInitialisation {
    foreach my $o (@option_list) {
        my $i = $o->{init};
        $i =~ s/X/*this, "$o->{name}"/;
        print "      $o->{name}$i,\n";
    }
}

sub generateSet {
    foreach my $o (@option_list) {
        if (exists($o->{set})) {
            print "    $o->{name}.$o->{set};\n";
        }
    }
}

sub generateSetDepend {
    foreach my $o (@option_list) {
        if (exists($o->{set}) && $o->{set} =~ /copyFrom\((.*)\)/) {
            print "    if (!$o->{name}.wasSet())\n";
            print "        $o->{name}.$o->{set};\n";
        }
    }
}

sub generateTim {
    my @import_list = sort {$a->{ttoffset} <=> $b->{ttoffset}} grep {exists($_->{ttoffset})} @option_list;
    my $cond = 0;
    my $cond_index = 0;
    my $sizevar = 'size';
    my $arrayvar = 'data';
    foreach (@import_list) {
        if ($_->{ttoffset} >= $hconfig_ranges[$cond_index]) {
            if ($cond) { print "    }\n"; }
            while ($cond_index <= $#hconfig_ranges && $_->{ttoffset} >= $hconfig_ranges[$cond_index]) {
                ++$cond_index
            }
            if ($cond_index > $#hconfig_ranges) {
                print STDERR "option index for $_->{name} out of range, update \@hconfig_ranges\n";
                return;
            }
            print "    if ($sizevar >= $hconfig_ranges[$cond_index]) {\n";
            $cond = 1;
        }
        my $c = $_->{ttimport};
        $c =~ s/\bX\b/$_->{name}/g;
        $c =~ s/\bOO\b/$_->{ttoffset}/g;
        $c =~ s/\bHH\b/$arrayvar/g;
        print "        $c; $_->{name}.setSource(src);\n";
    }
    if ($cond) { print "    }\n"; }
}

sub processFiles {
    foreach my $x (@_) {
        open CONF, "< $x" or die "can't open $x: $!";
        while(<CONF>) {
            if (/<option.*\bname="(.*?)"/) {
                my $opt = { name => $1 };
                if (/\btype="(.*?)"/) {
                    $opt->{type} = $1;
                }
                while (<CONF>) {
                    last if m|</option>|;
                    if (m|<([a-zA-Z0-9-_:]+)>(.*)</|) { $opt->{$1} = $2 }
                }
                addOption($opt);
            } else {
                processLine ();
            }
        }
        close CONF;
    }
}

sub processLine {
    if (/^CFDefine\((\S+),\s*\S+,\s*(\S+),\s*(\S+),\s*"(.*)",\s*(\S+),\s*(\S+),\s*(\S+)\s*\)/) {
	my ($key, $count, $type, $default, $array, $min, $max)
	    = ($1, $2, $3, $4, lc $5, $6, $7);
        doDefine ($key, $count, $type, $default, $array, $min, $max);
    } elsif(/^CFDefine4\((\S+),\s*\S+,\s*(\S+),\s*(\S+),\s*(\S+),\s*(\S+),\s*([^,]*),\s*([^,]*?),\s*"(.*)"\)/) {
        my ($key, $count, $type, $min, $max, $flags, $cond, $default)
	    = ($1, $2, $3, $4, $5, $6, $7, $8);
        my $array = ($flags =~ /ARRAYIZED/i) ? "true" : "false";
        doDefine ($key, $count, $type, $default, $array, $min, $max);
    } elsif(/^CFAlias\((\S+),\s*(\S+)\)/) {
        doAlias ($1, $2);
    } elsif(/^TCDefine\((\S+),\s*(\S+),\s*(\S+),\s*(\S+)\)/) {
        doTHost ($4, $1, $2, $3);
    } elsif(/^CFDefine/) {
	print STDERR "unmatched CFDefine: $_";
    }
}

sub doTHost {
    my ($key, $typ, $dim, $ofs) = @_;
    if (!defined($options_by_name{$key})) {
        print STDERR "must have PHost definition of '$key' before THost version\n";
        return;
    }
    my $o = $options_by_name{$key};

    # We either have the same type, or the PHost version has 11
    # entries and the THost version has one. No need to handle "*"
    # here.
    if (!exists($translation{$typ}{$dim})) {
        print STDERR "no type known for '$key'\n";
        return;
    }
    if (!exists($translation{$typ}{$dim}{import})) {
        print STDERR "no import known for '$key' of type $typ\[$dim]\n";
        return;
    }
    if ($translation{$typ}{$dim}{type} ne $o->{type}) {
        # different type
        if ($dim == 1) {
            $o->{ttimport} = $translation{$typ}{1}{import};
        } else {
            print STDERR "strange arrayisation of $key, $translation{$typ}{$dim} vs $o->{type}\n";
            return;
        }
    } else {
        # same type
        $o->{ttimport} = $translation{$typ}{$dim}{import};
    }
    $o->{ttoffset} = $ofs;
}

sub doDefine {
    my ($key, $count, $type, $default, $array, $min, $max) = @_;
    $count = replaceConstants($count);
    $min = replaceConstants($min);
    $max = replaceConstants($max);

    my $p;
    if (!defined($translation{$type})) {
        die "no translation for $key, type $type";
    }
    if (defined($translation{$type}{$count})) {
        $p = $translation{$type}{$count}
    } elsif (defined($translation{$type}{'*'})) {
        $p = $translation{$type}{'*'};
    } else {
        die "no translation for $key, type $type x $count";
    }

    my $init = $p->{init};
    my $asgn;
    if ($default =~ /^\$(.*)/) {
        $asgn = "copyFrom($1)";
    } elsif ($default =~ /,/ || $p->{type} =~ /String|Cost/) {
        $default =~ s/, +/,/g;
        $asgn = "set(\"$default\")";
    } else {
        $default =~ s/\b(yes|true)\b/1/gi;
        $default =~ s/\b(false|no)\b/0/gi;
        $default =~ s/\b(allies)\b/2 \/* Allies *\//gi;
        $default =~ s/\b(external)\b/3 \/* External *\//gi;
        $asgn = "set($default)";
    }

    addOption ({
                type => $p->{type},
                name => $key,
                init => $init,
                set => $asgn,
                "zztype" => $type . "[$count]",
                "zzrange" => "$min..$max"
               });
}

sub doAlias {
    my ($alias, $realname) = @_;
    addOption ({
                type => 'ConfigAlias',
                name => $alias,
                init => "(X, $realname)"
               });
}

sub replaceConstants {
    my $x = shift;
    $x eq 'PLANET_NR' and $x = 'NUM_PLANETS';
    $x eq 'SHIP_NR' and $x = 'NUM_SHIPS';
    $x;
}

sub addOption {
    foreach my $o (@_) {
        my $name = $o->{name};
        if (!exists $options_by_name{$name}) {
            $options_by_name{$name} = $o;
            push @option_list, $o;
        }
    }
}
