/**
  *  \file tools/c2mgrep.cc
  *  \brief "Message Grep" utility
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <vector>
#include <cstring>
#include "arch/platform.h"
#include "cpluslib/format.h"
#include "cpluslib/string.h"
#include "game/checksum.h"
#include "game/msg.h"
#include "game/racename.h"
#include "game/rst.h"
#include "game/struct.h"
#include "game/inbox.h"
#include "game/load.h"
#include "game/trn.h"
#include "io/io.h"
#include "io/memstream.h"
#include "io/streams.h"
#include "io/zip.h"
#include "util/console.h"
#include "util/except.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/recode.h"
#include "version.h"

#define VERSION PCC2_VERSION

static const char* progname;

/** Exit with error. */
static void
die(string_t str)
{
    console.write(LOG_ERROR, format("%s: %s", progname, str));
    std::exit(1);
}

/** Exit with help message. */
static void
help()
{
    console << format(_("PCC2 Message Search v%s - (c) 2011-2015 Stefan Reuther\n"), VERSION)
            << format(_("\n"
                        "Usage:\n"
                        "  %s [-h]\n"
                        "  %$0s [-opts] \"search string\" [-type] files...\n\n"
                        "Options:\n"
                        "  -c           Case-sensitive\n"
                        "  -C CHARSET   Select character set\n"
                        "\n"
                        "Type options apply to all subsequent file names:\n"
                        "  -r           Result files\n"
                        "  -t           Turn files\n"
                        "  -m           Message inbox (MDATA)\n"
                        "  -d           Dosplan outbox (MESS)\n"
                        "  -w           Winplan outbox (MESS35)\n"
                        "  -a           VPA database\n"
                        "  -A           Auto-detect (default)\n"
                        "  -z           Search in ZIP files (implies -A)\n"
                        "  -I           Ignore unknown files (default: warn)\n"
                        "\n"
                        "Report bugs to <Streu@gmx.de>\n"),
                      progname);
    console.flush();
    std::exit(0);
}

/************************** File Identification **************************/

static const char vpadbsig[] = "VPA Database\015\012\006";
static const size_t vpadbsiglen = sizeof(vpadbsig)-1;

enum FileType {
    UnknownFile,
    TurnFile,
    ResultFile,
    OutboxFile,
    Outbox35File,
    InboxFile,
    VpaDatabaseFile,
    ZipArchive
};

static bool
isTurn(const char* data, size_t size, long total)
{
    if (size <= TTurnHeader::size)
        return false;

    /* Validate ranges. 100000 is an estimate how many commands a turn file can
       possibly contain. MGREP 1.3d uses 5000, which is consistent with Host, but
       a turn can legally contain more commands. */
    /* FIXME: this does not recognize Taccom. Should it? */
    TTurnHeader hdr;
    getStructure(data, hdr);
    if (hdr.player <= 0 || hdr.player > 11)
        return false;
    if (hdr.num_commands < 0 || hdr.num_commands > 100000)
        return false;
    if (getChecksum(hdr.timestamp, sizeof(hdr.timestamp)) != hdr.time_checksum)
        return false;
    if (total < TTurnHeader::size || total - TTurnHeader::size < hdr.num_commands*4)
        return false;
    return true;
}

static bool
isResult(const char* data, size_t size, long total)
{
    if (size <= 8*4)
        return false;

    /* Check pointers in standard range */
    for (int i = 0; i < 8; ++i) {
        int32_t v = getInt32(data + 4*i);
        if (v < 20 || v > total)
            return false;
    }
    return true;
}

static bool
isOutboxFile(const char* data, size_t size, long total)
{
    if (size < 2)
        return false;

    /* Check count, and that the header fits in the file completely. */
    uint16_t nmess = getUint16(data);
    long minsize = nmess*long(TOutgoingMessageHeader::size) + 2;
    if (minsize > total)
        return false;

    /* Figure out how many headers fit in our preview */
    size_t max = (size - 2) / TOutgoingMessageHeader::size;
    if (nmess > max)
        nmess = max;

    /* Check all headers. Length and position must be reasonable,
       sender numbers must agree, receiver must be in range. */
    int sender = 0;
    for (uint16_t i = 0; i < nmess; ++i) {
        TOutgoingMessageHeader omh;
        getStructure(data + 2 + TOutgoingMessageHeader::size*i, omh);
        if (omh.address < minsize || omh.address > total)
            return false;
        if (omh.length < 0 || omh.length > 10000)
            return false;
        if (omh.to <= 0 || omh.to > 12)
            return false;
        if (sender == 0) {
            if (omh.from <= 0 || omh.from > 11)
                return false;
            sender = omh.from;
        } else {
            if (omh.from != sender)
                return false;
        }
    }
    return true;
}

static bool
isOutbox35File(const char* data, size_t size, long total)
{
    /* File consists of a word with a count, 18 garbage bytes, and 13
       flag bytes, where the flag bytes are actually part of a
       repeating structure. Check that count and flag bytes are valid.
       It makes no sense to check the repeating structure, as our
       preview will probably be too small anyway. */
    int16_t count = getInt16(data);
    if (size < 2+18+13)
        return false;
    if (count < 0 || count > 10000)
        return false;
    if (total < 100)
        return false;
    for (int i = 0; i < 13; ++i)
        if (data[20+i] != '0' && data[20+i] != '1')
            return false;
    return true;
}

static bool
isInbox(const char* data, size_t size, long total)
{
    if (size < 2)
        return false;

    /* Check count, and that the header fits in the file completely. */
    uint16_t nmess = getUint16(data);
    long minsize = nmess*long(TIncomingMessageHeader::size) + 2;
    if (minsize > total)
        return false;

    /* Figure out how many headers fit in our preview */
    size_t max = (size - 2) / TIncomingMessageHeader::size;
    if (nmess > max)
        nmess = max;

    /* Check all headers. Length and position must be reasonable. */
    for (uint16_t i = 0; i < nmess; ++i) {
        TIncomingMessageHeader imh;
        getStructure(data + 2 + TIncomingMessageHeader::size*i, imh);
        if (imh.address < minsize || imh.address > total)
            return false;
        if (imh.length < 0 || imh.length > 10000)
            return false;
    }
    return true;
}

static bool
isVpaDatabase(const char* data, size_t size, long total)
{
    if (total < 128 || size < vpadbsiglen)
        return false;
    return (std::memcmp(data, vpadbsig, vpadbsiglen) == 0);
}

static bool
isZipFile(const char* data, size_t size, long total)
{
    return total >= 100  /* 31 bytes local header, 47 bytes central header, 22 bytes end */
        && size >= 4
        && data[0] == 'P'
        && data[1] == 'K'
        && data[2] == 3
        && data[3] == 4;
}

static FileType
identifyFile(Stream& s)
{
    /* Read first bytes of the file */
    char tmp[512];
    size_t n   = s.read(tmp, sizeof(tmp));
    long total = s.getSize();
    s.seek(0);

    /* Apply all checkers, most reliable first */
    if (isZipFile(tmp, n, total))
        return ZipArchive;
    if (isVpaDatabase(tmp, n, total))
        return VpaDatabaseFile;
    if (isTurn(tmp, n, total))
        return TurnFile;
    if (isResult(tmp, n, total))
        return ResultFile;
    if (isOutboxFile(tmp, n, total))
        return OutboxFile;
    if (isOutbox35File(tmp, n, total))
        return Outbox35File;
    if (isInbox(tmp, n, total))
        return InboxFile;
    return UnknownFile;
}

/********************************* Search ********************************/

namespace {
    struct Message {
        string_t text;
        string_t header;
        string_t file;
        int turn;
        int index;

        string_t query;
        bool optCaseSense;

        Message()
            : turn(0), index(0), optCaseSense(false)
            { }
        void search();
    };

    class OutboxSearch : public GOutboxReader {
        Message& m;
     public:
        OutboxSearch(Message& m)
            : m(m)
            { }
        void addMessage(string_t message, GPlayerSet receivers);
    };

    struct Job {
        string_t query;
        bool optCaseSense;
        bool optAllowZip;
        bool optWarnUnknown;
        FileType optFileType;

        Job()
            : query(), optCaseSense(false), optAllowZip(false), optWarnUnknown(true), optFileType(UnknownFile)
            { }
    };
}

static void searchStream(Ptr<Stream> file, string_t fname, const Job& job);

void
Message::search()
{
    string_t t = text;
    if (!optCaseSense)
        t = strUCase(t);

    if (query.size() == 0 || t.find(query) != t.npos) {
        /* Divider */
        if (!index)
            console << "--- Message ---\n";
        else if (file.empty())
            console << "--- Message " << index << "---\n";
        else
            console << "--- Message " << index << " (" << file << ") ---\n";

        /* Header */
        console << header;
        if (turn)
            console << "TURN: " << turn << "\n";

        /* Body */
        console << text << "\n";
    }
}

void
OutboxSearch::addMessage(string_t message, GPlayerSet receivers)
{
    ++m.index;
    m.text   = message;
    m.header = format("TO: %s\n", formatPlayerSet(receivers));
    m.search();
}

static void
searchInbox(Message& m, Stream& s)
{
    /* This is simple, but inefficient */
    GInbox inbox;
    inbox.loadInbox(s, 0);
    for (int i = 0, e = inbox.getCount(); i < e; ++i) {
        m.text = inbox.getText(i);
        m.index = i+1;
        m.search();
    }
}

static void
searchOutbox(Message& m, Stream& s)
{
    OutboxSearch o(m);
    m.index = 0;
    o.loadOutbox(s);
}

static void
searchOutbox35(Message& m, Stream& s)
{
    OutboxSearch o(m);
    m.index = 0;
    o.loadOutbox35(s);
}

static void
searchResult(Message& m, Stream& s)
{
    GResultFile rst(s);

    // extract turn number
    TResultGen rg;
    s.seek(rst.getSectionOffset(GResultFile::sec_Gen));
    getStructureT(s, rg);
    m.turn = rg.turn;

    // messages
    s.seek(rst.getSectionOffset(GResultFile::sec_Messages));
    searchInbox(m, s);
}

static void
searchTurn(Message& m, Stream& s)
{
    GTurnfile trn(s);
    m.turn = trn.tryGetTurnNr();
    for (int i = 0, e = trn.getNumCommands(); i < e; ++i) {
        if (trn.getCommandCode(i) == tcm_SendMessage) {
            const char* data = trn.getCommandData(i);
            m.text = getMessageFromArray(data+4, trn.getCommandId(i), false);
            m.header = format("FROM: Player %d\nTO: Player %d\n", trn.getPlayer(), getUint16(data+2));
            ++m.index;
            m.search();
        }
    }
}

static void
searchVPA(Message& m, Stream& s)
{
    // Signature: 15 bytes
    // Turn header: 116 bytes
    // Chunk header: 10 bytes
    enum {
        TURN_SIZE  = 116,
        CHUNK_SIZE = 10
    };
    char buf[TURN_SIZE];
    s.readT(buf, vpadbsiglen);
    if (std::memcmp(buf, vpadbsig, vpadbsiglen) != 0)
        throw FileFormatException(s, _("File is missing required signature"));

    while (s.read(buf, TURN_SIZE) == TURN_SIZE) {
        if (std::memcmp(buf, "TURN", 4) != 0)
            throw FileFormatException(s, _("Invalid file block"));
        m.turn = getInt16(buf+8);
        int32_t turnSize = getInt32(buf+4);
        while (turnSize >= CHUNK_SIZE) {
            // Read chunk header and advance positions
            s.readT(buf, CHUNK_SIZE);
            int32_t chunkSize = getInt32(buf+4);
            turnSize -= CHUNK_SIZE;
            turnSize -= chunkSize;
            int32_t endPos = s.getPos() + chunkSize;

            // Check chunk
            if (std::memcmp(buf, "IMSG", 4) == 0) {
                // Build virtual inbox file and read that;
                // include the "count" word from the header
                SubStream sub(s);
                sub.setRange(s.getPos()-2, chunkSize+2);
                sub.seek(0);
                searchInbox(m, sub);
            } else if (std::memcmp(buf, "OMSG", 4) == 0) {
                // Again, build virtual outbox file and read that.
                SubStream sub(s);
                sub.setRange(s.getPos()-2, chunkSize+2);
                sub.seek(0);
                searchOutbox(m, sub);
            }

            // Update file pointer
            s.seek(endPos);
        }
        if (turnSize != 0)
            throw FileFormatException(s, _("Invalid file block"));
    }
}

static void
searchZip(Ptr<Stream> file, string_t fname, const Job& job)
{
    // Construct sub-job as modified version of existing job
    Job subjob(job);
    subjob.optAllowZip = false;
    subjob.optWarnUnknown = false;

    // Iterate through zip file
    Ptr<ZipFile> zip = ZipFile::open(file, ZipFile::KeepPaths);
    Ptr<Enumeration<Ptr<DirectoryEntry> > > zents = zip->getEntries();
    while (zents->hasMoreElements()) {
        Ptr<DirectoryEntry> zent = zents->getNextElement();
        if (zent->getFileType() == DirectoryEntry::FT_FILE) {
            Ptr<Stream> zstream = zent->openFile(Stream::C_READ);
            if (!zstream->canDo(Stream::C_SEEK)) {
                // Entry cannot seek. Read into memory.
                Buffer<char> content;
                zstream->readBuffer(content);
                Ptr<MemoryStream> ms(new MemoryStream());
                ms->swapBuffer(content);
                zstream = ms;
            }
            searchStream(zstream, fname + "(" + zent->getName() + ")", subjob);
        }
    }
}

static void
searchStream(Ptr<Stream> file, string_t fname, const Job& job)
{
    FileType type = job.optFileType == UnknownFile ? identifyFile(*file) : job.optFileType;

    Message m;
    m.file = fname;
    m.optCaseSense = job.optCaseSense;
    m.query = job.optCaseSense ? job.query : strUCase(job.query);

    switch (type) {
     case TurnFile:
        searchTurn(m, *file);
        break;
     case ResultFile:
        searchResult(m, *file);
        break;
     case OutboxFile:
        searchOutbox(m, *file);
        break;
     case Outbox35File:
        searchOutbox35(m, *file);
        break;
     case InboxFile:
        searchInbox(m, *file);
        break;
     case VpaDatabaseFile:
        searchVPA(m, *file);
        break;
     case ZipArchive:
        if (job.optAllowZip) {
            searchZip(file, fname, job);
        } else {
            console.write(LOG_ERROR, format(_("%s: compressed file"), fname));
        }
        break;
     case UnknownFile:
        if (job.optWarnUnknown) {
            console.write(LOG_ERROR, format(_("%s: unknown file format"), fname));
        }
        break;
    }
}
          
static void
searchFile(const char* fname, const Job& job)
{
    try {
        Ptr<Stream> file = Stream::openFile(fname, Stream::C_READ);
        searchStream(file, fname, job);
    }
    catch (FileProblemException& ex) {
        console.write(LOG_ERROR, format("%s: %s", ex.getFileName(), ex.what()));
    }
    catch (std::exception& ex) {
        console.write(LOG_ERROR, format("%s: %s", fname, ex.what()));
    }
}

static const ModuleInitializer*const mgrep_modules[] = {
    mod_base_console,
    mod_racenames,
    0
};

int
main(int /*argc*/, char** argv)
{
    /* Arguments */
    bool hadSearchString = false;
    bool hadFiles = false;
    Job job;

    /* Initialize */
    performInitialisation(OnStartup, mgrep_modules);
    argv = Platform::convertArgumentsToUtf8(argv);
    progname = argv[0];

    /* Parse and search */
    while (const char* arg = *++argv) {
        if (*arg == '-') {
            while (char c = *++arg) {
                switch (c) {
                 case 'h':
                    help();
                    break;
                 case 'C':
                    if (! *++argv)
                        die(_("option `-C' needs an argument (the character set)"));
                    {
                        CharacterSet cs(CharacterSet::getCharsetByName(*argv));
                        if (!cs.isValid())
                            die(_("the specified character set is not known"));
                        setGameCharacterSet(cs);
                    }
                 case 'r':
                    job.optFileType = ResultFile;
                    break;
                 case 't':
                    job.optFileType = TurnFile;
                    break;
                 case 'm':
                    job.optFileType = InboxFile;
                    break;
                 case 'd':
                    job.optFileType = OutboxFile;
                    break;
                 case 'w':
                    job.optFileType = Outbox35File;
                    break;
                 case 'a':
                    job.optFileType = VpaDatabaseFile;
                    break;
                 case 'A':
                    job.optFileType = UnknownFile;
                    break;
                 case 'z':
                    job.optAllowZip = true;
                    job.optFileType = UnknownFile;
                    break;
                 case 'c':
                    job.optCaseSense = true;
                    break;
                 case 'I':
                    job.optWarnUnknown = false;
                    break;
                 case 'n':      // disable mbox format
                 case 'f':      // enable mbox format
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        } else if (!hadSearchString) {
            job.query = arg;
            hadSearchString = true;
        } else {
            searchFile(arg, job);
            hadFiles = true;
        }
    }

    if (!hadSearchString) {
        die(format(_("no search string specified. Use `%s -h' for help"), progname));
    }
    if (!hadFiles) {
        die(format(_("no file name specified. Use `%s -h' for help"), progname));
    }
    return 0;
}
